import json
from typing import Tuple, List, Optional, Union

from requests_oauthlib import OAuth2Session

from pymfy.api.devices.category import Category
from pymfy.api.model import Command, Site, Device

BASE_URL = 'https://api.somfy.com/api/v1'

SOMFY_OAUTH = 'https://accounts.somfy.com/oauth/oauth/v2/auth'
SOMFY_TOKEN = 'https://accounts.somfy.com/oauth/oauth/v2/token'
SOMFY_REFRESH = 'https://accounts.somfy.com/oauth/oauth/v2/token'


class SomfyApi:
    __slots__ = '_oauth', 'client_id', 'client_secret', 'cache_path', '_token'

    def __init__(self, client_id: str, client_secret: str, redirect_uri: str,
                 cache_path: str = None):

        self.client_id = client_id
        self.client_secret = client_secret
        self.cache_path = cache_path
        self.token = None

        extra = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }

        def token_setter(token):
            self.token = token

        self._oauth = OAuth2Session(client_id=client_id,
                                    token=self.token,
                                    redirect_uri=redirect_uri,
                                    auto_refresh_kwargs=extra,
                                    auto_refresh_url=SOMFY_REFRESH,
                                    token_updater=token_setter)

    def get_authorization_url(self) -> Tuple[str, str]:
        return self._oauth.authorization_url(SOMFY_OAUTH)

    def request_token(self, authorization_response: Optional[str] = None,
                      code: Optional[str] = None) -> None:
        """Generic method for fetching a Somfy access token.

        :param authorization_response: Authorization response URL, the callback
                                       URL of the request back to you.
        :param code: Authorization code
        """
        self.token = self._oauth.fetch_token(
            SOMFY_TOKEN,
            authorization_response=authorization_response,
            code=code,
            client_secret=self.client_secret)

    @property
    def token(self) -> str:
        token = self._token
        if self.cache_path:
            try:
                with open(self.cache_path, 'r') as cache:
                    token = json.loads(cache.read())
            except IOError:
                pass
        return token

    @token.setter
    def token(self, token: str) -> None:
        self._token = token
        if self.cache_path and token:
            with open(self.cache_path, 'w') as cache:
                cache.write(json.dumps(token))

    def get_sites(self) -> List[Site]:
        r = self._oauth.get(BASE_URL + '/site')
        r.raise_for_status()
        return [Site(s) for s in r.json()]

    def get_site(self, site_id: str) -> Site:
        r = self._oauth.get(BASE_URL + '/site/' + site_id)
        r.raise_for_status()
        return Site(r.json())

    def send_command(self, device_id: str,
                     command: Union[Command, str]) -> str:
        if isinstance(command, str):
            command = Command(command)
        r = self._oauth.post(BASE_URL + '/device/' + device_id + '/exec',
                             json=command)
        r.raise_for_status()
        return r.json().get('job_id')

    def get_devices(self, site_id: Optional[str] = None,
                    category: Optional[Category] = None) -> List[Device]:
        site_ids = [s.id for s in self.get_sites()] if site_id is None else [
            site_id]
        devices = []
        for site_id in site_ids:
            r = self._oauth.get(BASE_URL + '/site/' + site_id + "/device")
            r.raise_for_status()
            devices += [Device(d) for d in r.json() if
                        category is None or category.value in Device(
                            d).categories]
        return devices

    def get_device(self, device_id: str) -> Device:
        r = self._oauth.get(BASE_URL + '/device/' + device_id)
        r.raise_for_status()
        return Device(r.json())



from pymfy.api.devices.base import SomfyDevice
from pymfy.api.model import Command, Parameter


class Thermostat(SomfyDevice):
    """Class to represent a thermostat."""

    def get_ambient_temperature(self) -> float:
        return self.get_state('ambient_temperature')

    def get_humidity(self) -> float:
        return self.get_state('humidity')

    def get_battery(self) -> int:
        return self.get_state('battery')

    def get_hvac_state(self) -> str:
        return self.get_state('hvac_state')

    def get_regulation_state(self) -> str:
        return self.get_state('regulation_state')

    def get_target_mode(self) -> str:
        return self.get_state('target_mode')

    def get_target_temperature(self) -> int:
        return self.get_state('target_temperature')

    def get_derogation_end_date(self) -> int:
        return self.get_state('derogation.end_date')

    def get_derogation_start_date(self) -> int:
        return self.get_state('derogation.start_date')

    def get_at_home_temperature(self) -> int:
        return self.get_state('at_home_temperature')

    def get_away_temperature(self) -> int:
        return self.get_state('away_temperature')

    def get_night_temperature(self) -> int:
        return self.get_state('night_temperature')

    def get_frost_protection_temperature(self) -> int:
        return self.get_state('frost_protection_temperature')

    def set_target(self, target_mode: str, target_temperature: int,
                   duration: int, duration_type: str) -> None:
        parameters = [Parameter('target_mode', target_mode),
                      Parameter('target_temperature', target_temperature),
                      Parameter('duration', duration),
                      Parameter('duration_type', duration_type)]
        command = Command('set_target', parameters)
        self.send_command(command)

    def cancel_target(self) -> None:
        self.send_command(Command('cancel_target'))

    def set_at_home_temperature(self, temperature: int) -> None:
        parameter = Parameter('at_home_temperature', temperature)
        command = Command('set_at_home_temperature', parameter)
        self.send_command(command)

    def set_away_temperature(self, temperature: int) -> None:
        parameter = Parameter('away_temperature', temperature)
        command = Command('set_away_temperature', parameter)
        self.send_command(command)

    def set_night_temperature(self, temperature: int) -> None:
        parameter = Parameter('night_temperature', temperature)
        command = Command('set_night_temperature', parameter)
        self.send_command(command)

    def set_frost_protection_temperature(self, temperature: int) -> None:
        parameter = Parameter('frost_protection_temperature', temperature)
        command = Command('set_frost_protection_temperature', parameter)
        self.send_command(command)



from typing import Dict, Union, List, Any


class Site(dict):
    __slots__ = 'id', 'label'

    def __init__(self, json: Dict[str, str]):
        self.id = json.get('id')
        self.label = json.get('label')
        dict.__init__(self, id=self.id, label=self.label)


class Device(dict):
    __slots__ = 'id', 'name', 'type', 'site_id', 'states', 'capabilities', \
                'categories'

    def __init__(self, json: Dict[str, Any]):
        self.id = json.get('id')
        self.name = json.get('name')
        self.type = json.get('type')
        self.site_id = json.get('site_id')
        self.categories = json.get('categories')
        self.states = [State(s) for s in json.get('states')]
        self.capabilities = [Capability(c) for c in json.get('capabilities')]
        dict.__init__(self, id=self.id, name=self.name, type=self.type,
                      site_id=self.site_id, categories=self.categories,
                      states=self.states, capabilities=self.capabilities)


class State(dict):
    __slots__ = 'name', 'value', 'type'

    def __init__(self, json: Dict[str, str]):
        self.name = json.get('name')
        self.value = json.get('value')
        self.type = json.get('type')
        dict.__init__(self, name=self.name, value=self.value, type=self.type)


class Capability(dict):
    __slots__ = 'name', 'parameters'

    def __init__(self, json: Dict[str, str]):
        self.name = json.get('name')
        self.parameters = [ParameterDescription(p) for p in
                           json.get('parameters')]
        dict.__init__(self, name=self.name, parameters=self.parameters)


class ParameterDescription(dict):
    __slots__ = 'name', 'type'

    def __init__(self, json: Dict[str, str]):
        self.name = json.get('name')
        self.type = json.get('type')
        dict.__init__(self, name=self.name, type=self.type)


class Command(dict):
    __slots__ = 'name', 'parameters'

    def __init__(self, name: str, parameters: Union[
            List[Dict[str, str]], Dict[str, str]] = None):
        if parameters is None:
            parameters = []
        if not isinstance(parameters, list):
            parameters = [parameters]
        self.name = name
        self.parameters = parameters
        dict.__init__(self, name=name, parameters=parameters)


class Parameter(dict):
    __slots__ = 'name', 'value'

    def __init__(self, name: str, value: Union[str, int]):
        self.name = name
        self.value = value
        dict.__init__(self, name=name, value=value)



from typing import Optional

from pymfy.api.devices.base import SomfyDevice
from pymfy.api.model import Command, Parameter


class RollerShutter(SomfyDevice):
    """Class to represent a roller shutter."""

    def get_position(self) -> int:
        return self.get_state('position') or 0

    def set_position(self, value: int,
                     low_speed: Optional[int] = False) -> None:
        command_name = 'position_low_speed' if low_speed else 'position'
        command = Command(command_name, Parameter('position', value))
        self.send_command(command)

    def close(self) -> None:
        self.send_command(Command('close'))

    def open(self) -> None:
        self.send_command(Command('open'))

    def stop(self) -> None:
        self.send_command(Command('stop'))

    def identify(self) -> None:
        self.send_command(Command('identify'))

    def is_closed(self) -> bool:
        return self.get_position() == 100



from typing import Any

from pymfy.api.devices.base import SomfyDevice
from pymfy.api.model import Command


class CameraProtect(SomfyDevice):
    """Class to represent a camera"""

    def close_shutter(self) -> None:
        self.send_command(Command('close'))

    def open_shutter(self) -> None:
        self.send_command(Command('open'))

    # TODO Waiting documentation to know what's the returned type
    def get_shutter_position(self) -> Any:
        return self.get_state('shutter_position')
